# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

from unittest import mock

import hikari
import pytest
from hikari import traits

import tanjun


class TestAutocompleteOption:
    def test_is_focused_property(self):
        mock_raw_option = mock.Mock()
        option = tanjun.context.AutocompleteOption(mock.Mock(), mock_raw_option)

        assert option.is_focused is mock_raw_option.is_focused


class TestAutocompleteContext:
    @pytest.fixture()
    def mock_client(self) -> tanjun.Client:
        return mock.AsyncMock(tanjun.Client)

    @pytest.fixture()
    def mock_interaction(self) -> hikari.AutocompleteInteraction:
        return mock.Mock(options=[mock.Mock(is_focused=True)])

    @pytest.fixture()
    def context(
        self, mock_client: tanjun.Client, mock_interaction: hikari.AutocompleteInteraction
    ) -> tanjun.context.AutocompleteContext:
        return tanjun.context.AutocompleteContext(mock_client, mock_interaction)

    def test_author_property(
        self, context: tanjun.context.AutocompleteContext, mock_interaction: hikari.AutocompleteInteraction
    ):
        assert context.author is mock_interaction.user

    def test_channel_id_property(
        self, context: tanjun.context.AutocompleteContext, mock_interaction: hikari.AutocompleteInteraction
    ):
        assert context.channel_id is mock_interaction.channel_id

    def test_cache_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.cache is mock_client.cache

    def test_client_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.client is mock_client

    def test_created_at_property(
        self, context: tanjun.context.AutocompleteContext, mock_interaction: hikari.AutocompleteInteraction
    ):
        assert context.created_at is mock_interaction.created_at

    def test_events_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.events is mock_client.events

    def test_focused_property(self):
        focused_option = mock.Mock(is_focused=True)
        mock_interaction = mock.Mock(options=[mock.Mock(is_focused=False), focused_option, mock.Mock(is_focused=False)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction)

        assert isinstance(context.focused, tanjun.context.AutocompleteOption)
        assert context.focused._resolved is mock_interaction.resolved
        assert context.focused._option is focused_option

    def test_guild_id_property(
        self, context: tanjun.context.AutocompleteContext, mock_interaction: hikari.AutocompleteInteraction
    ):
        assert context.guild_id is mock_interaction.guild_id

    def test_member_property(
        self, context: tanjun.context.AutocompleteContext, mock_interaction: hikari.AutocompleteInteraction
    ):
        assert context.member is mock_interaction.member

    def test_server_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.server is mock_client.server

    def test_rest_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.rest is mock_client.rest

    def test_shard_property(
        self,
        context: tanjun.context.AutocompleteContext,
        mock_client: mock.Mock,
        mock_interaction: hikari.AutocompleteInteraction,
    ):
        mock_shard = mock.Mock()
        mock_client.shards = mock.MagicMock(spec=traits.ShardAware, shard_count=5, shards={2: mock_shard})
        mock_interaction.guild_id = hikari.Snowflake(123321123312)

        assert context.shard is mock_shard

    def test_shard_property_when_dm(
        self,
        context: tanjun.context.AutocompleteContext,
        mock_client: mock.Mock,
        mock_interaction: hikari.AutocompleteInteraction,
    ):
        mock_shard = mock.Mock()
        mock_client.shards = mock.Mock(shards={0: mock_shard})
        mock_interaction.guild_id = None

        assert context.shard is mock_shard

    def test_shard_property_when_no_shards(self, context: tanjun.context.AutocompleteContext):
        context._client = mock.Mock(shards=None)

        assert context.shard is None

    def test_shards_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.shards is mock_client.shards

    def test_voice_property(self, context: tanjun.context.AutocompleteContext, mock_client: tanjun.Client):
        assert context.voice is mock_client.voice

    def test_has_responded_property(self, context: tanjun.context.AutocompleteContext):
        assert context.has_responded is False

    def test_interaction_property(
        self, context: tanjun.context.AutocompleteContext, mock_interaction: hikari.AutocompleteInteraction
    ):
        assert context.interaction is mock_interaction

    def test_options_property(self):
        mock_option_1 = mock.Mock()
        mock_option_2 = mock.Mock()
        mock_option_3 = mock.Mock()
        context = tanjun.context.AutocompleteContext(
            mock.Mock(tanjun.Client), mock.Mock(options=[mock_option_1, mock_option_2, mock_option_3])
        )

        assert len(context.options) == 3
        for (name, option), raw_option in zip(context.options.items(), [mock_option_1, mock_option_2, mock_option_3]):
            assert isinstance(option, tanjun.context.AutocompleteOption)
            assert name is option.name
            assert option._option is raw_option

    # def test_options_property_for_top_level_command(self, mock_client: mock.Mock):
    #     mock_option_1 = mock.Mock()
    #     mock_option_1.name = "hi"
    #     mock_option_2 = mock.Mock()
    #     mock_option_2.name = "bye"
    #     context = tanjun.context.SlashContext(
    #         mock_client,
    #         mock.Mock(),
    #         mock.Mock(type=hikari.OptionType.SUB_COMMAND, options=[mock_option_1, mock_option_2]),
    #         command=mock.Mock(),
    #         component=mock.Mock(),
    #     )

    #     assert len(context.options) == 2
    #     assert context.options["hi"].type is mock_option_1.type
    #     assert context.options["hi"].value is mock_option_1.value
    #     assert context.options["hi"].name is mock_option_1.name
    #     assert isinstance(context.options["hi"], tanjun.context.SlashOption)

    #     assert context.options["bye"].type is mock_option_2.type
    #     assert context.options["bye"].value is mock_option_2.value
    #     assert context.options["bye"].name is mock_option_2.name
    #     assert isinstance(context.options["bye"], tanjun.context.SlashOption)

    # def test_options_property_for_command_group(self, mock_client: mock.Mock):
    #     mock_option_1 = mock.Mock()
    #     mock_option_1.name = "kachow"
    #     mock_option_2 = mock.Mock()
    #     mock_option_2.name = "nyaa"
    #     group_option = mock.Mock(type=hikari.OptionType.SUB_COMMAND, options=[mock_option_1, mock_option_2])
    #     context = tanjun.context.SlashContext(
    #         mock_client,
    #         mock.Mock(),
    #         mock.Mock(type=hikari.OptionType.SUB_COMMAND_GROUP, options=[group_option]),
    #         command=mock.Mock(),
    #         component=mock.Mock(),
    #     )

    #     assert len(context.options) == 2
    #     assert context.options["kachow"].type is mock_option_1.type
    #     assert context.options["kachow"].value is mock_option_1.value
    #     assert context.options["kachow"].name is mock_option_1.name
    #     assert isinstance(context.options["kachow"], tanjun.context.SlashOption)

    #     assert context.options["nyaa"].type is mock_option_2.type
    #     assert context.options["nyaa"].value is mock_option_2.value
    #     assert context.options["nyaa"].name is mock_option_2.name
    #     assert isinstance(context.options["nyaa"], tanjun.context.SlashOption)

    # @pytest.mark.parametrize("raw_options", [None, []])
    # def test_options_property_for_command_group_with_no_sub_option(
    #     self, mock_client: mock.Mock, raw_options: typing.Optional[list[hikari.OptionType]]
    # ):
    #     group_option = mock.Mock(type=hikari.OptionType.SUB_COMMAND, options=raw_options)
    #     context = tanjun.context.SlashContext(
    #         mock_client,
    #         mock.Mock(),
    #         mock.Mock(type=hikari.OptionType.SUB_COMMAND_GROUP, options=[group_option]),
    #         command=mock.Mock(),
    #         component=mock.Mock(),
    #     )

    #     assert context.options == {}

    # def test_options_property_for_sub_command_group(self, mock_client: mock.Mock):
    #     mock_option_1 = mock.Mock()
    #     mock_option_1.name = "meow"
    #     mock_option_2 = mock.Mock()
    #     mock_option_2.name = "nya"
    #     sub_group_option = mock.Mock(type=hikari.OptionType.SUB_COMMAND, options=[mock_option_1, mock_option_2])
    #     group_option = mock.Mock(type=hikari.OptionType.SUB_COMMAND_GROUP, options=[sub_group_option])
    #     context = tanjun.context.SlashContext(
    #         mock_client,
    #         mock.Mock(),
    #         mock.Mock(type=hikari.OptionType.SUB_COMMAND_GROUP, options=[group_option]),
    #         command=mock.Mock(),
    #         component=mock.Mock(),
    #     )

    #     assert len(context.options) == 2
    #     assert context.options["meow"].type is mock_option_1.type
    #     assert context.options["meow"].value is mock_option_1.value
    #     assert context.options["meow"].name is mock_option_1.name
    #     assert isinstance(context.options["meow"], tanjun.context.SlashOption)

    #     assert context.options["nya"].type is mock_option_2.type
    #     assert context.options["nya"].value is mock_option_2.value
    #     assert context.options["nya"].name is mock_option_2.name
    #     assert isinstance(context.options["nya"], tanjun.context.SlashOption)

    @pytest.mark.asyncio()
    async def test_fetch_channel(self, context: tanjun.context.AutocompleteContext, mock_interaction: mock.Mock):
        mock_interaction.fetch_channel = mock.AsyncMock()

        result = await context.fetch_channel()

        assert result is mock_interaction.fetch_channel.return_value
        mock_interaction.fetch_channel.assert_awaited_once_with()

    @pytest.mark.asyncio()
    async def test_fetch_guild(self, context: tanjun.context.AutocompleteContext, mock_interaction: mock.Mock):
        mock_interaction.fetch_guild = mock.AsyncMock()

        result = await context.fetch_guild()

        assert result is mock_interaction.fetch_guild.return_value
        mock_interaction.fetch_guild.assert_awaited_once_with()

    def test_get_channel(self, context: tanjun.context.AutocompleteContext, mock_interaction: mock.Mock):
        result = context.get_channel()

        assert result is mock_interaction.get_channel.return_value
        mock_interaction.get_channel.assert_called_once_with()

    def test_get_guild(self, context: tanjun.context.AutocompleteContext, mock_interaction: mock.Mock):
        result = context.get_guild()

        assert result is mock_interaction.get_guild.return_value
        mock_interaction.get_guild.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_set_choices(self):
        mock_interaction = mock.AsyncMock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction)
        assert context.has_responded is False

        result = await context.set_choices({"hi": "bye"}, ok="45")

        assert result is None
        assert context.has_responded is True
        mock_interaction.create_response.assert_awaited_once_with(
            [hikari.CommandChoice(name="hi", value="bye"), hikari.CommandChoice(name="ok", value="45")]
        )

    @pytest.mark.asyncio()
    async def test_set_choices_when_no_options(self):
        mock_interaction = mock.AsyncMock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction)

        result = await context.set_choices()

        assert result is None
        mock_interaction.create_response.assert_awaited_once_with([])

    @pytest.mark.asyncio()
    async def test_set_choices_when_iterable(self):
        mock_interaction = mock.AsyncMock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction)

        result = await context.set_choices([("echo", "delta"), ("hi", "bye"), ("goof", "meow")])

        assert result is None
        mock_interaction.create_response.assert_awaited_once_with(
            [
                hikari.CommandChoice(name="echo", value="delta"),
                hikari.CommandChoice(name="hi", value="bye"),
                hikari.CommandChoice(name="goof", value="meow"),
            ]
        )

    @pytest.mark.asyncio()
    async def test_set_choices_when_future_based(self):
        mock_future = mock.Mock()
        mock_interaction = mock.Mock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction, future=mock_future)

        result = await context.set_choices({"meow": "neko", "woof": "borf"}, bark="moo")

        assert result is None
        mock_future.set_result.assert_called_once_with(mock_interaction.build_response.return_value)
        mock_interaction.build_response.assert_called_once_with(
            [
                hikari.CommandChoice(name="meow", value="neko"),
                hikari.CommandChoice(name="woof", value="borf"),
                hikari.CommandChoice(name="bark", value="moo"),
            ]
        )

    @pytest.mark.asyncio()
    async def test_set_choices_when_future_based_and_iterable(self):
        mock_future = mock.Mock()
        mock_interaction = mock.Mock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction, future=mock_future)
        assert context.has_responded is False

        result = await context.set_choices([("japan", "weeb"), ("england", "idk")], usa="lol")

        assert result is None
        mock_future.set_result.assert_called_once_with(mock_interaction.build_response.return_value)
        mock_interaction.build_response.assert_called_once_with(
            [
                hikari.CommandChoice(name="japan", value="weeb"),
                hikari.CommandChoice(name="england", value="idk"),
                hikari.CommandChoice(name="usa", value="lol"),
            ]
        )
        assert context.has_responded is True

    @pytest.mark.asyncio()
    async def test_set_choices_when_future_based_and_no_options(self):
        mock_future = mock.Mock()
        mock_interaction = mock.Mock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction, future=mock_future)

        result = await context.set_choices()

        assert result is None
        mock_future.set_result.assert_called_once_with(mock_interaction.build_response.return_value)
        mock_interaction.build_response.assert_called_once_with([])

    @pytest.mark.asyncio()
    async def test_set_choices_when_too_many_options(self):
        mock_interaction = mock.AsyncMock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction)

        with pytest.raises(ValueError, match="Cannot set more than 25 choices"):
            await context.set_choices({str(v): "a" for v in range(0, 23)}, foo="ok", bar="as", baz="43")

        mock_interaction.create_response.assert_not_called()

    @pytest.mark.asyncio()
    async def test_set_choices_when_has_responded(self):
        mock_interaction = mock.AsyncMock(options=[mock.Mock(is_focused=True)])
        context = tanjun.context.AutocompleteContext(mock.Mock(tanjun.Client), mock_interaction)
        await context.set_choices()
        mock_interaction.create_response.reset_mock()

        with pytest.raises(RuntimeError, match="Cannot set choices after responding"):
            await context.set_choices()

        mock_interaction.create_response.assert_not_called()



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# This leads to too many false-positives around mocks.

import typing
from unittest import mock

import hikari
import pytest

import tanjun


@pytest.fixture()
def command() -> tanjun.abc.ExecutableCommand[typing.Any]:
    command_ = mock.MagicMock(tanjun.abc.ExecutableCommand)
    command_.add_check.return_value = command_
    return command_


@pytest.fixture()
def context() -> tanjun.abc.Context:
    return mock.MagicMock(tanjun.abc.Context)


class TestOwnerCheck:
    @pytest.mark.asyncio()
    async def test(self):
        mock_dependency = mock.AsyncMock()
        mock_dependency.check_ownership.return_value = True
        mock_context = mock.Mock()
        check = tanjun.checks.OwnerCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, mock_dependency)

        assert result is True
        mock_dependency.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)

    @pytest.mark.asyncio()
    async def test_when_false(self):
        mock_dependency = mock.AsyncMock()
        mock_dependency.check_ownership.return_value = False
        mock_context = mock.Mock()
        check = tanjun.checks.OwnerCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, mock_dependency)

        assert result is False
        mock_dependency.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)

    @pytest.mark.asyncio()
    async def test_when_false_and_error_message(self):
        mock_dependency = mock.AsyncMock()
        mock_dependency.check_ownership.return_value = False
        mock_context = mock.Mock()
        check = tanjun.checks.OwnerCheck(error_message="aye", halt_execution=False)

        with pytest.raises(tanjun.errors.CommandError, match="aye"):
            await check(mock_context, mock_dependency)

        mock_dependency.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)

    @pytest.mark.asyncio()
    async def test_when_false_and_halt_execution(self):
        mock_dependency = mock.AsyncMock()
        mock_dependency.check_ownership.return_value = False
        mock_context = mock.Mock()
        check = tanjun.checks.OwnerCheck(error_message=None, halt_execution=True)

        with pytest.raises(tanjun.errors.HaltExecution):
            await check(mock_context, mock_dependency)

        mock_dependency.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)


class TestNsfwCheck:
    @pytest.mark.asyncio()
    async def test_when_is_dm(self):
        mock_context = mock.Mock(guild_id=None)
        mock_cache = mock.AsyncMock()
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.cache.get_guild_channel.assert_not_called()
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_not_called()

    @pytest.mark.asyncio()
    async def test(self):
        mock_context = mock.Mock()
        mock_context.cache.get_guild_channel.return_value.is_nsfw = True
        mock_cache = mock.AsyncMock()
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_not_called()

    @pytest.mark.asyncio()
    async def test_when_async_cache_raises_not_found(self):
        mock_context = mock.Mock(cache=None, rest=mock.AsyncMock())
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.EntryNotFound
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        with pytest.raises(tanjun.dependencies.EntryNotFound):
            await check(mock_context, channel_cache=mock_cache)

        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_called_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_cache_bound_and_async_cache_hit(self):
        mock_context = mock.Mock(cache=None, rest=mock.AsyncMock())
        mock_cache = mock.AsyncMock()
        mock_cache.get.return_value.is_nsfw = True
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_called_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_found_in_cache_and_async_cache_hit(self):
        mock_context = mock.Mock(rest=mock.AsyncMock())
        mock_context.cache.get_guild_channel.return_value = None
        mock_cache = mock.AsyncMock()
        mock_cache.get.return_value.is_nsfw = None
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is False
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_called_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_cache_bound(self):
        mock_context = mock.Mock(cache=None, rest=mock.AsyncMock())
        mock_context.rest.fetch_channel.return_value = mock.Mock(hikari.GuildChannel, is_nsfw=True)
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=None)

        assert result is True
        mock_context.rest.fetch_channel.assert_awaited_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_found_in_cache(self):
        mock_context = mock.Mock(rest=mock.AsyncMock())
        mock_context.cache.get_guild_channel.return_value = None
        mock_context.rest.fetch_channel.return_value = mock.Mock(hikari.GuildChannel, is_nsfw=True)
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.CacheMissError
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_awaited_once_with(mock_context.channel_id)
        mock_cache.get.assert_awaited_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_false(self):
        mock_context = mock.Mock()
        mock_context.cache.get_guild_channel.return_value.is_nsfw = None
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=None)

        assert result is False
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()

    @pytest.mark.asyncio()
    async def test_when_false_and_error_message(self):
        mock_context = mock.Mock()
        mock_context.cache.get_guild_channel.return_value.is_nsfw = False
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.CacheMissError
        check = tanjun.checks.NsfwCheck(error_message="meow me", halt_execution=False)

        with pytest.raises(tanjun.errors.CommandError, match="meow me"):
            await check(mock_context, channel_cache=mock_cache)

        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_not_called()

    @pytest.mark.asyncio()
    async def test_when_false_and_halt_execution(self):
        mock_context = mock.Mock(rest=mock.AsyncMock())
        mock_context.cache.get_guild_channel.return_value = None
        mock_context.rest.fetch_channel.return_value = mock.Mock(hikari.GuildChannel, is_nsfw=False)
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.CacheMissError
        check = tanjun.checks.NsfwCheck(error_message=None, halt_execution=True)

        with pytest.raises(tanjun.errors.HaltExecution):
            await check(mock_context, channel_cache=mock_cache)

        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_awaited_once_with(mock_context.channel_id)
        mock_cache.get.assert_awaited_once_with(mock_context.channel_id)


class TestSfwCheck:
    @pytest.mark.asyncio()
    async def test_when_is_dm(self):
        mock_context = mock.Mock(guild_id=None)
        mock_cache = mock.AsyncMock()
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.cache.get_guild_channel.assert_not_called()
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_not_called()

    @pytest.mark.asyncio()
    async def test(self):
        mock_context = mock.Mock()
        mock_context.cache.get_guild_channel.return_value.is_nsfw = False
        mock_cache = mock.AsyncMock()
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_not_called()

    @pytest.mark.asyncio()
    async def test_when_not_cache_bound_and_async_cache_hit(self):
        mock_context = mock.Mock(cache=None, rest=mock.AsyncMock())
        mock_cache = mock.AsyncMock()
        mock_cache.get.return_value.is_nsfw = False
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_called_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_found_in_cache_and_async_cache_hit(self):
        mock_context = mock.Mock(rest=mock.AsyncMock())
        mock_context.cache.get_guild_channel.return_value = None
        mock_cache = mock.AsyncMock()
        mock_cache.get.return_value.is_nsfw = None
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is True
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_called_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_cache_bound(self):
        mock_context = mock.Mock(cache=None, rest=mock.AsyncMock())
        mock_context.rest.fetch_channel.return_value = mock.Mock(hikari.GuildChannel, is_nsfw=True)
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=None)

        assert result is False
        mock_context.rest.fetch_channel.assert_awaited_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_not_found_in_cache(self):
        mock_context = mock.Mock(rest=mock.AsyncMock())
        mock_context.cache.get_guild_channel.return_value = None
        mock_context.rest.fetch_channel.return_value = mock.Mock(hikari.GuildChannel, is_nsfw=True)
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.CacheMissError
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=mock_cache)

        assert result is False
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_awaited_once_with(mock_context.channel_id)
        mock_cache.get.assert_awaited_once_with(mock_context.channel_id)

    @pytest.mark.asyncio()
    async def test_when_is_nsfw(self):
        mock_context = mock.Mock()
        mock_context.cache.get_guild_channel.return_value.is_nsfw = True
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=False)

        result = await check(mock_context, channel_cache=None)

        assert result is False
        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()

    @pytest.mark.asyncio()
    async def test_when_is_nsfw_and_error_message(self):
        mock_context = mock.Mock()
        mock_context.cache.get_guild_channel.return_value.is_nsfw = True
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.CacheMissError
        check = tanjun.checks.SfwCheck(error_message="meow me", halt_execution=False)

        with pytest.raises(tanjun.errors.CommandError, match="meow me"):
            await check(mock_context, channel_cache=mock_cache)

        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_not_called()
        mock_cache.get.assert_not_called()

    @pytest.mark.asyncio()
    async def test_when_is_nsfw_and_halt_execution(self):
        mock_context = mock.Mock(rest=mock.AsyncMock())
        mock_context.cache.get_guild_channel.return_value = None
        mock_context.rest.fetch_channel.return_value = mock.Mock(hikari.GuildChannel, is_nsfw=True)
        mock_cache = mock.AsyncMock()
        mock_cache.get.side_effect = tanjun.dependencies.CacheMissError
        check = tanjun.checks.SfwCheck(error_message=None, halt_execution=True)

        with pytest.raises(tanjun.errors.HaltExecution):
            await check(mock_context, channel_cache=mock_cache)

        mock_context.cache.get_guild_channel.assert_called_once_with(mock_context.channel_id)
        mock_context.rest.fetch_channel.assert_awaited_once_with(mock_context.channel_id)
        mock_cache.get.assert_awaited_once_with(mock_context.channel_id)


class TestDmCheck:
    def test_for_dm(self):
        assert tanjun.checks.DmCheck()(mock.Mock(guild_id=None)) is True

    def test_for_guild(self):
        assert tanjun.checks.DmCheck(halt_execution=False, error_message=None)(mock.Mock(guild_id=3123)) is False

    def test_for_guild_when_halt_execution(self):
        with pytest.raises(tanjun.HaltExecution):
            assert tanjun.checks.DmCheck(halt_execution=True, error_message=None)(mock.Mock(guild_id=3123))

    def test_for_guild_when_error_message(self):
        with pytest.raises(tanjun.CommandError, match="message"):
            assert tanjun.checks.DmCheck(halt_execution=False, error_message="message")(mock.Mock(guild_id=3123))


class TestGuildCheck:
    def test_for_guild(self):
        assert tanjun.checks.GuildCheck()(mock.Mock(guild_id=123123)) is True

    def test_for_dm(self):
        assert tanjun.checks.GuildCheck(halt_execution=False, error_message=None)(mock.Mock(guild_id=None)) is False

    def test_for_dm_when_halt_execution(self):
        with pytest.raises(tanjun.HaltExecution):
            tanjun.checks.GuildCheck(halt_execution=True, error_message=None)(mock.Mock(guild_id=None))

    def test_for_dm_when_error_message(self):
        with pytest.raises(tanjun.CommandError, match="hi"):
            tanjun.checks.GuildCheck(halt_execution=False, error_message="hi")(mock.Mock(guild_id=None))


@pytest.mark.skip(reason="Not Implemented")
class TestAuthorPermissionCheck:
    ...


@pytest.mark.skip(reason="Not Implemented")
class TestOwnPermissionCheck:
    ...


def test_with_dm_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "DmCheck") as dm_check:
        assert tanjun.checks.with_dm_check(command) is command

        command.add_check.assert_called_once_with(dm_check.return_value)
        dm_check.assert_called_once_with(halt_execution=False, error_message="Command can only be used in DMs")


def test_with_dm_check_with_keyword_arguments(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "DmCheck") as dm_check:
        assert tanjun.checks.with_dm_check(halt_execution=True, error_message="message")(command) is command

        command.add_check.assert_called_once_with(dm_check.return_value)
        dm_check.assert_called_once_with(halt_execution=True, error_message="message")


def test_with_guild_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "GuildCheck") as guild_check:
        assert tanjun.checks.with_guild_check(command) is command

        command.add_check.assert_called_once_with(guild_check.return_value)
        guild_check.assert_called_once_with(
            halt_execution=False, error_message="Command can only be used in guild channels"
        )


def test_with_guild_check_with_keyword_arguments(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "GuildCheck") as guild_check:
        assert tanjun.checks.with_guild_check(halt_execution=True, error_message="eee")(command) is command

        command.add_check.assert_called_once_with(guild_check.return_value)
        guild_check.assert_called_once_with(halt_execution=True, error_message="eee")


def test_with_nsfw_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "NsfwCheck", return_value=mock.AsyncMock()) as nsfw_check:
        assert tanjun.checks.with_nsfw_check(command) is command

        command.add_check.assert_called_once_with(nsfw_check.return_value)
        nsfw_check.assert_called_once_with(
            halt_execution=False, error_message="Command can only be used in NSFW channels"
        )


def test_with_nsfw_check_with_keyword_arguments(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "NsfwCheck", return_value=mock.AsyncMock()) as nsfw_check:
        assert tanjun.checks.with_nsfw_check(halt_execution=True, error_message="banned!!!")(command) is command

        command.add_check.assert_called_once_with(nsfw_check.return_value)
        nsfw_check.assert_called_once_with(halt_execution=True, error_message="banned!!!")


def test_with_sfw_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "SfwCheck", return_value=mock.AsyncMock()) as sfw_check:
        assert tanjun.checks.with_sfw_check(command) is command

        command.add_check.assert_called_once_with(sfw_check.return_value)
        sfw_check.assert_called_once_with(
            halt_execution=False, error_message="Command can only be used in SFW channels"
        )


def test_with_sfw_check_with_keyword_arguments(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "SfwCheck", return_value=mock.AsyncMock()) as sfw_check:
        assert tanjun.checks.with_sfw_check(halt_execution=True, error_message="bango")(command) is command

        command.add_check.assert_called_once_with(sfw_check.return_value)
        sfw_check.assert_called_once_with(halt_execution=True, error_message="bango")


def test_with_owner_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "OwnerCheck") as owner_check:
        assert tanjun.checks.with_owner_check(command) is command

        command.add_check.assert_called_once_with(owner_check.return_value)
        owner_check.assert_called_once_with(halt_execution=False, error_message="Only bot owners can use this command")


def test_with_owner_check_with_keyword_arguments(command: mock.Mock):
    mock_check = object()
    with mock.patch.object(tanjun.checks, "OwnerCheck", return_value=mock_check) as owner_check:
        result = tanjun.checks.with_owner_check(
            halt_execution=True,
            error_message="dango",
        )(command)
        assert result is command

        command.add_check.assert_called_once_with(owner_check.return_value)
        owner_check.assert_called_once_with(halt_execution=True, error_message="dango")


def test_with_author_permission_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "AuthorPermissionCheck") as author_permission_check:
        assert (
            tanjun.checks.with_author_permission_check(435213, halt_execution=True, error_message="bye")(command)
            is command
        )

        command.add_check.assert_called_once_with(author_permission_check.return_value)
        author_permission_check.assert_called_once_with(435213, halt_execution=True, error_message="bye")


def test_with_own_permission_check(command: mock.Mock):
    with mock.patch.object(tanjun.checks, "OwnPermissionCheck") as own_permission_check:
        assert (
            tanjun.checks.with_own_permission_check(5412312, halt_execution=True, error_message="hi")(command)
            is command
        )

        command.add_check.assert_called_once_with(own_permission_check.return_value)
        own_permission_check.assert_called_once_with(5412312, halt_execution=True, error_message="hi")


def test_with_check(command: mock.Mock):
    mock_check = mock.Mock()

    result = tanjun.checks.with_check(mock_check)(command)

    assert result is command
    command.add_check.assert_called_once_with(mock_check)


@pytest.mark.asyncio()
async def test_all_checks():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(return_value=True)
    check = tanjun.checks.all_checks(mock_check_1, mock_check_2, mock_check_3)

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_all_checks_when_check_raises():
    class MockError(Exception):
        ...

    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[True, MockError])
    check = tanjun.checks.all_checks(mock_check_1, mock_check_2, mock_check_3)

    with pytest.raises(MockError):
        await check(mock_context)

    mock_context.call_with_async_di.assert_has_awaits(
        [mock.call(mock_check_1, mock_context), mock.call(mock_check_2, mock_context)]
    )


@pytest.mark.asyncio()
async def test_all_checks_when_first_fails():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(return_value=False)
    check = tanjun.checks.all_checks(mock_check_1, mock_check_2, mock_check_3)

    result = await check(mock_context)

    assert result is False
    mock_context.call_with_async_di.assert_awaited_once_with(mock_check_1, mock_context)


@pytest.mark.asyncio()
async def test_all_checks_when_last_fails():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[True, True, False])
    check = tanjun.checks.all_checks(mock_check_1, mock_check_2, mock_check_3)

    result = await check(mock_context)

    assert result is False
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_all_checks_when_any_check_fails():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_check_4 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[True, False, True, True])
    check = tanjun.checks.all_checks(mock_check_1, mock_check_2, mock_check_3, mock_check_4)

    result = await check(mock_context)

    assert result is False
    mock_context.call_with_async_di.assert_has_awaits(
        [mock.call(mock_check_1, mock_context), mock.call(mock_check_2, mock_context)]
    )


def test_with_all_checks():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_command = mock.Mock()

    with mock.patch.object(tanjun.checks, "all_checks") as all_checks:
        result = tanjun.with_all_checks(mock_check_1, mock_check_2, mock_check_3)(mock_command)

    assert result is mock_command.add_check.return_value
    mock_command.add_check.assert_called_once_with(all_checks.return_value)
    all_checks.assert_called_once_with(mock_check_1, mock_check_2, mock_check_3)


@pytest.mark.asyncio()
async def test_any_checks_when_first_check_passes():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(return_value=True)
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message="hi")

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_awaited_once_with(mock_check_1, mock_context)


@pytest.mark.asyncio()
async def test_any_checks_when_last_check_passes():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.FailedCheck, True])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message="hi")

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_check_passes():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_check_4 = mock.Mock()
    mock_check_5 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.FailedCheck, False, True])
    check = tanjun.checks.any_checks(
        mock_check_1, mock_check_2, mock_check_3, mock_check_4, mock_check_5, error_message="hi"
    )

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
            mock.call(mock_check_4, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_all_fail():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.FailedCheck, False])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None)

    result = await check(mock_context)

    assert result is False
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_all_fail_and_halt_execution():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, False, tanjun.FailedCheck])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None, halt_execution=True)

    with pytest.raises(tanjun.HaltExecution):
        await check(mock_context)

    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_all_fail_and_error_message():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[tanjun.FailedCheck, False, False])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message="aye")

    with pytest.raises(tanjun.CommandError, match="aye"):
        await check(mock_context)

    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_generic_unsuppressed_error_raised():
    class MockError(Exception):
        ...

    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, MockError])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None)

    with pytest.raises(MockError):
        await check(mock_context)

    mock_context.call_with_async_di.assert_has_awaits(
        [mock.call(mock_check_1, mock_context), mock.call(mock_check_2, mock_context)]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_generic_error_suppressed():
    class MockError(Exception):
        ...

    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, MockError, True])
    check = tanjun.checks.any_checks(
        mock_check_1, mock_check_2, mock_check_3, error_message=None, suppress=(MockError,)
    )

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_halt_execution_not_suppressed():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.HaltExecution])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None, suppress=())

    with pytest.raises(tanjun.HaltExecution):
        await check(mock_context)

    mock_context.call_with_async_di.assert_has_awaits(
        [mock.call(mock_check_1, mock_context), mock.call(mock_check_2, mock_context)]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_halt_execution_suppressed():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.HaltExecution, True])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None)

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_command_error_not_suppressed():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.CommandError("bye")])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None, suppress=())

    with pytest.raises(tanjun.CommandError, match="bye"):
        await check(mock_context)

    mock_context.call_with_async_di.assert_has_awaits(
        [mock.call(mock_check_1, mock_context), mock.call(mock_check_2, mock_context)]
    )


@pytest.mark.asyncio()
async def test_any_checks_when_command_error_suppressed():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_context = mock.Mock()
    mock_context.call_with_async_di = mock.AsyncMock(side_effect=[False, tanjun.CommandError("bye"), True])
    check = tanjun.checks.any_checks(mock_check_1, mock_check_2, mock_check_3, error_message=None)

    result = await check(mock_context)

    assert result is True
    mock_context.call_with_async_di.assert_has_awaits(
        [
            mock.call(mock_check_1, mock_context),
            mock.call(mock_check_2, mock_context),
            mock.call(mock_check_3, mock_context),
        ]
    )


def test_with_any_checks():
    mock_check_1 = mock.Mock()
    mock_check_2 = mock.Mock()
    mock_check_3 = mock.Mock()
    mock_command = mock.Mock()

    class MockError(Exception):
        ...

    with mock.patch.object(tanjun.checks, "any_checks") as any_checks:
        result = tanjun.checks.with_any_checks(
            mock_check_1,
            mock_check_2,
            mock_check_3,
            suppress=(MockError,),
            error_message="yay catgirls",
            halt_execution=True,
        )(mock_command)

    assert result is mock_command.add_check.return_value
    mock_command.add_check.assert_called_once_with(any_checks.return_value)
    any_checks.assert_called_once_with(
        mock_check_1,
        mock_check_2,
        mock_check_3,
        error_message="yay catgirls",
        suppress=(MockError,),
        halt_execution=True,
    )



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Standard base mplementation for Tanjun's command objects."""
from __future__ import annotations

__all__: list[str] = ["PartialCommand"]

import copy
import typing
from collections import abc as collections

from .. import abc
from .. import components

if typing.TYPE_CHECKING:
    _CheckSigT = typing.TypeVar("_CheckSigT", bound=abc.CheckSig)
    _PartialCommandT = typing.TypeVar("_PartialCommandT", bound="PartialCommand[typing.Any]")


_ContextT = typing.TypeVar("_ContextT", bound="abc.Context")


class PartialCommand(abc.ExecutableCommand[_ContextT], components.AbstractComponentLoader):
    """Base class for the standard ExecutableCommand implementations."""

    __slots__ = ("_checks", "_component", "_hooks", "_metadata")

    def __init__(self) -> None:
        self._checks: list[abc.CheckSig] = []
        self._component: typing.Optional[abc.Component] = None
        self._hooks: typing.Optional[abc.Hooks[_ContextT]] = None
        self._metadata: dict[typing.Any, typing.Any] = {}

    @property
    def checks(self) -> collections.Collection[abc.CheckSig]:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        return self._checks.copy()

    @property
    def component(self) -> typing.Optional[abc.Component]:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        return self._component

    @property
    def hooks(self) -> typing.Optional[abc.Hooks[_ContextT]]:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        return self._hooks

    @property
    def metadata(self) -> collections.MutableMapping[typing.Any, typing.Any]:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        return self._metadata

    def copy(self: _PartialCommandT, *, _new: bool = True) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        if not _new:
            self._checks = [copy.copy(check) for check in self._checks]
            self._hooks = self._hooks.copy() if self._hooks else None
            self._metadata = self._metadata.copy()
            return self

        return copy.copy(self).copy(_new=False)

    def set_hooks(self: _PartialCommandT, hooks: typing.Optional[abc.Hooks[_ContextT]], /) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        self._hooks = hooks
        return self

    def set_metadata(self: _PartialCommandT, key: typing.Any, value: typing.Any, /) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        self._metadata[key] = value
        return self

    def add_check(self: _PartialCommandT, check: abc.CheckSig, /) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        if check not in self._checks:
            self._checks.append(check)

        return self

    def remove_check(self: _PartialCommandT, check: abc.CheckSig, /) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        self._checks.remove(check)
        return self

    def with_check(self, check: _CheckSigT, /) -> _CheckSigT:
        self.add_check(check)
        return check

    def bind_client(self: _PartialCommandT, client: abc.Client, /) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        return self

    def bind_component(self: _PartialCommandT, component: abc.Component, /) -> _PartialCommandT:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        self._component = component
        return self



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.
import asyncio
import contextlib
import datetime
import time
import typing
from unittest import mock

import alluka
import hikari
import pytest

import tanjun
from tanjun.context import base as base_context


@pytest.mark.parametrize(
    ("resource_type", "mock_ctx", "expected"),
    [
        (tanjun.BucketResource.USER, mock.Mock(author=mock.Mock(id=123321)), 123321),
        (tanjun.BucketResource.CHANNEL, mock.Mock(channel_id=43433123), 43433123),
        (tanjun.BucketResource.GUILD, mock.Mock(guild_id=65123), 65123),
        (tanjun.BucketResource.GUILD, mock.Mock(guild_id=None, channel_id=611223), 611223),
    ],
)
@pytest.mark.asyncio()
async def test__get_ctx_target(resource_type: tanjun.BucketResource, mock_ctx: tanjun.abc.Context, expected: int):
    assert await tanjun.dependencies.limiters._get_ctx_target(mock_ctx, resource_type) == expected


@pytest.mark.parametrize(
    ("channel", "result"),
    [(mock.Mock(parent_id=None, id=123), 123), (mock.Mock(parent_id=54123123, id=123321), 123321)],
)
@pytest.mark.asyncio()
async def test__get_ctx_target_when_parent_channel_and_cache_result(channel: mock.Mock, result: int):
    mock_context = mock.Mock()
    mock_context.get_channel.return_value = channel

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.PARENT_CHANNEL)

    assert result == result
    mock_context.get_channel.assert_called_once_with()
    mock_context.fetch_channel.assert_not_called()
    mock_context.get_type_dependency.assert_not_called()


@pytest.mark.parametrize(
    ("channel", "result"),
    [(mock.Mock(parent_id=None, id=123), 123), (mock.Mock(parent_id=54123123, id=123321), 123321)],
)
@pytest.mark.asyncio()
async def test__get_ctx_target_when_parent_channel_when_async_cache_returns_channel(channel: mock.Mock, result: int):
    mock_cache = mock.AsyncMock()
    mock_cache.get.return_value = channel
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.get_channel.return_value = None
    mock_context.get_type_dependency.return_value = mock_cache

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.PARENT_CHANNEL)

    assert result == result
    mock_context.get_channel.assert_called_once_with()
    mock_cache.get.assert_awaited_once_with(mock_context.channel_id, default=None)
    mock_context.fetch_channel.assert_not_called()
    mock_context.get_type_dependency.assert_called_once_with(tanjun.dependencies.SfCache[hikari.GuildChannel])


@pytest.mark.parametrize(
    ("channel", "result"),
    [
        (mock.Mock(hikari.TextableGuildChannel, parent_id=None, id=123), 123),
        (mock.Mock(hikari.TextableGuildChannel, parent_id=54123123, id=123321), 123321),
    ],
)
@pytest.mark.asyncio()
async def test__get_ctx_target_when_parent_channel_when_async_cache_returns_none_falls_back_to_rest(
    channel: mock.Mock, result: int
):
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.get_channel.return_value = None
    mock_context.fetch_channel = mock.AsyncMock(return_value=channel)
    mock_cache = mock.AsyncMock()
    mock_cache.get.return_value = None
    mock_context.get_type_dependency.return_value = mock_cache

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.PARENT_CHANNEL)

    assert result == result
    mock_context.get_channel.assert_called_once_with()
    mock_context.fetch_channel.assert_awaited_once()
    mock_context.get_type_dependency.assert_called_once_with(tanjun.dependencies.SfCache[hikari.GuildChannel])
    mock_cache.get.assert_awaited_once_with(mock_context.channel_id, default=None)


@pytest.mark.parametrize(
    ("channel", "result"),
    [
        (mock.Mock(hikari.TextableGuildChannel, parent_id=None, id=123), 123),
        (mock.Mock(hikari.TextableGuildChannel, parent_id=54123123, id=123321), 123321),
    ],
)
@pytest.mark.asyncio()
async def test__get_ctx_target_when_parent_channel_and_no_async_cache_falls_back_to_rest(
    channel: mock.Mock, result: int
):
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.get_channel.return_value = None
    mock_context.fetch_channel = mock.AsyncMock(return_value=channel)
    mock_context.get_type_dependency.return_value = alluka.abc.UNDEFINED

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.PARENT_CHANNEL)

    assert result == result
    mock_context.get_channel.assert_called_once_with()
    mock_context.fetch_channel.assert_awaited_once()
    mock_context.get_type_dependency.assert_called_once_with(tanjun.dependencies.SfCache[hikari.GuildChannel])


@pytest.mark.asyncio()
async def test__get_ctx_target_when_parent_channel_and_dm_bound():
    mock_context = mock.Mock(guild_id=None)

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.PARENT_CHANNEL)

    assert result is mock_context.channel_id


@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role():
    mock_roles = [
        mock.Mock(id=123321, position=555),
        mock.Mock(id=42123, position=56),
        mock.Mock(id=4234, position=333),
        mock.Mock(id=4354, position=0),
        mock.Mock(id=4123, position=11),
    ]
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.member.role_ids = [123, 312]
    mock_context.member.get_roles = mock.Mock(return_value=mock_roles)
    mock_context.member.fetch_roles = mock.AsyncMock()

    assert await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE) == 123321

    mock_context.member.get_roles.assert_called_once_with()
    mock_context.member.fetch_roles.assert_not_called()
    mock_context.get_type_dependency.assert_not_called()


@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role_and_async_cache():
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.member.role_ids = [674345, 123876, 7643, 9999999]
    mock_context.member.get_roles = mock.Mock(return_value=[])
    mock_cache = mock.AsyncMock()
    mock_cache.get.side_effect = [
        mock.Mock(position=42),
        mock.Mock(id=994949, position=7634),
        tanjun.dependencies.EntryNotFound,
        mock.Mock(position=23),
    ]
    mock_context.get_type_dependency.return_value = mock_cache

    assert await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE) == 994949

    mock_context.member.get_roles.assert_called_once_with()
    mock_context.member.fetch_roles.assert_not_called()
    mock_context.get_type_dependency.assert_called_once_with(tanjun.dependencies.SfCache[hikari.Role])
    mock_cache.get.assert_has_awaits([mock.call(674345), mock.call(123876), mock.call(7643), mock.call(9999999)])


@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role_falls_back_to_rest_when_async_cache_raises_cache_miss():
    mock_roles = [
        mock.Mock(id=123321, position=42),
        mock.Mock(id=123322, position=43),
        mock.Mock(id=431, position=6969),
        mock.Mock(id=111, position=0),
        mock.Mock(id=4123, position=6959),
    ]
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.member.role_ids = [123, 312, 654]
    mock_context.member.get_roles = mock.Mock(return_value=[])
    mock_context.member.fetch_roles = mock.AsyncMock(return_value=mock_roles)
    mock_cache = mock.AsyncMock()
    mock_cache.get.side_effect = [mock.Mock(), tanjun.dependencies.CacheMissError]
    mock_context.get_type_dependency.return_value = mock_cache

    assert await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE) == 431

    mock_context.member.get_roles.assert_called_once_with()
    mock_context.member.fetch_roles.assert_awaited_once_with()
    mock_context.get_type_dependency.assert_called_once_with(tanjun.dependencies.SfCache[hikari.Role])
    mock_cache.get.assert_has_awaits([mock.call(123), mock.call(312)])


@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role_falls_back_to_rest():
    mock_roles = [
        mock.Mock(id=123321, position=42),
        mock.Mock(id=123322, position=43),
        mock.Mock(id=431, position=6969),
        mock.Mock(id=111, position=0),
        mock.Mock(id=4123, position=6959),
    ]
    mock_context = mock.Mock(base_context.BaseContext)
    mock_context.member.role_ids = [123, 312]
    mock_context.member.get_roles = mock.Mock(return_value=[])
    mock_context.member.fetch_roles = mock.AsyncMock(return_value=mock_roles)
    mock_context.get_type_dependency.return_value = alluka.abc.UNDEFINED

    assert await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE) == 431

    mock_context.member.get_roles.assert_called_once_with()
    mock_context.member.fetch_roles.assert_awaited_once_with()
    mock_context.get_type_dependency.assert_called_once_with(tanjun.dependencies.SfCache[hikari.Role])


@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role_when_dm_bound():
    mock_context = mock.Mock(guild_id=None)

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE)

    assert result is mock_context.channel_id


@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role_when_no_member_in_guild():
    mock_context = mock.Mock(guild_id=654124, member=None)

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE)

    assert result == 654124
    mock_context.get_type_dependency.assert_not_called()


@pytest.mark.parametrize("role_ids", [[123321], []])
@pytest.mark.asyncio()
async def test__get_ctx_target_when_top_role_when_no_roles_or_only_1_role(role_ids: list[int]):
    mock_context = mock.Mock(guild_id=123312)
    mock_context.member.role_ids = role_ids

    result = await tanjun.dependencies.limiters._get_ctx_target(mock_context, tanjun.BucketResource.TOP_ROLE)

    assert result == 123312
    mock_context.get_type_dependency.assert_not_called()


@pytest.mark.asyncio()
async def test__get_ctx_target_when_unexpected_type():
    with pytest.raises(ValueError, match="Unexpected type 1"):
        await tanjun.dependencies.limiters._get_ctx_target(mock.Mock(), tanjun.BucketResource.MEMBER)


class Test_Cooldown:
    def test_has_expired_property(self):
        with mock.patch.object(time, "monotonic", side_effect=[69.0, 72.0]):
            cooldown = tanjun.dependencies.limiters._Cooldown(limit=1, reset_after=60.0)

            assert cooldown.has_expired() is False

    def test_has_expired_property_when_has_expired(self):
        with mock.patch.object(time, "monotonic", side_effect=[69.0, 96.0]):
            cooldown = tanjun.dependencies.limiters._Cooldown(limit=1, reset_after=26.5)

            assert cooldown.has_expired() is True

    def test_increment(self):
        with mock.patch.object(time, "monotonic", side_effect=[50.0, 55.0]):
            cooldown = tanjun.dependencies.limiters._Cooldown(limit=5, reset_after=69.420)
            cooldown.counter = 2

            cooldown.increment()

            assert cooldown.counter == 3
            assert cooldown.resets_at == 119.420

    def test_increment_when_counter_is_0(self):
        with mock.patch.object(time, "monotonic", side_effect=[419.0, 420.0]):
            cooldown = tanjun.dependencies.limiters._Cooldown(limit=5, reset_after=69.00)

            cooldown.increment()

            assert cooldown.counter == 1
            assert cooldown.resets_at == 489.00

    def test_increment_when_counter_is_at_limit(self):
        with mock.patch.object(time, "monotonic", side_effect=[419.0, 420.0]):
            cooldown = tanjun.dependencies.limiters._Cooldown(limit=5, reset_after=69.00)
            cooldown.counter = 5

            cooldown.increment()

            assert cooldown.counter == 5
            assert cooldown.resets_at == 488.00

    def test_increment_when_passed_reset_after(self):
        with mock.patch.object(time, "monotonic", side_effect=[419.0, 422.5]):
            cooldown = tanjun.dependencies.limiters._Cooldown(limit=5, reset_after=2.0)
            cooldown.counter = 2

            cooldown.increment()

            assert cooldown.counter == 1
            assert cooldown.resets_at == 424.5

    def test_increment_when_limit_is_negeative_1(self):
        cooldown = tanjun.dependencies.limiters._Cooldown(limit=-1, reset_after=-1)
        cooldown.resets_at = 54345543
        cooldown.counter = 123

        with mock.patch.object(time, "monotonic") as monotonic:
            result = cooldown.increment()

            monotonic.assert_not_called()

        assert result is cooldown
        assert cooldown.counter == 123
        assert cooldown.resets_at == 54345543

    def test_must_wait_for(self):
        with mock.patch.object(time, "monotonic", side_effect=[419.0, 422.0]):
            cooldown = tanjun.dependencies.limiters._Cooldown(reset_after=10.5, limit=3)
            cooldown.counter = 3

            assert cooldown.must_wait_for() == 7.5

    def test_must_wait_for_when_resource_limit_not_hit(self):
        cooldown = tanjun.dependencies.limiters._Cooldown(reset_after=2.0, limit=3)
        cooldown.counter = 2

        assert cooldown.must_wait_for() is None

    def test_must_wait_for_when_reset_after_reached(self):
        with mock.patch.object(time, "monotonic", side_effect=[419.0, 425.5]):
            cooldown = tanjun.dependencies.limiters._Cooldown(reset_after=5.0, limit=3)
            cooldown.counter = 3

            assert cooldown.must_wait_for() is None

    def test_must_wait_for_when_limit_is_negative_1(self):
        cooldown = tanjun.dependencies.limiters._Cooldown(reset_after=64, limit=-1)
        cooldown.counter = 55

        with mock.patch.object(time, "monotonic") as monotonic:
            assert cooldown.must_wait_for() is None

            monotonic.assert_not_called()


class Test_FlatResource:
    @pytest.mark.asyncio()
    async def test_try_into_inner(self):
        mock_resource_maker = mock.Mock()
        mock_resource = mock.Mock()
        bucket = tanjun.dependencies.limiters._FlatResource(tanjun.BucketResource.USER, mock_resource_maker)
        bucket.mapping[hikari.Snowflake(321123)] = mock_resource
        mock_context = mock.Mock()
        mock_context.author.id = hikari.Snowflake(321123)

        result = await bucket.try_into_inner(mock_context)

        assert result is mock_resource
        mock_resource_maker.assert_not_called()
        assert bucket.mapping[hikari.Snowflake(321123)] is mock_resource

    @pytest.mark.asyncio()
    async def test_try_into_inner_when_resource_doesnt_exist(self):
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._FlatResource(tanjun.BucketResource.USER, mock_resource_maker)
        mock_context = mock.Mock()
        mock_context.author.id = hikari.Snowflake(123321)

        result = await bucket.try_into_inner(mock_context)

        assert result is None
        mock_resource_maker.assert_not_called()
        assert hikari.Snowflake(123321) not in bucket.mapping

    @pytest.mark.asyncio()
    async def test_into_inner(self):
        mock_resource_maker = mock.Mock()
        mock_resource = mock.Mock()
        bucket = tanjun.dependencies.limiters._FlatResource(tanjun.BucketResource.USER, mock_resource_maker)
        bucket.mapping[hikari.Snowflake(3333)] = mock_resource
        mock_context = mock.Mock()
        mock_context.author.id = hikari.Snowflake(3333)

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource
        mock_resource_maker.assert_not_called()
        assert bucket.mapping[hikari.Snowflake(3333)] is mock_resource

    @pytest.mark.asyncio()
    async def test_into_inner_creates_new_resource(self):
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._FlatResource(tanjun.BucketResource.USER, mock_resource_maker)
        mock_context = mock.Mock()
        mock_context.author.id = hikari.Snowflake(123)

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource_maker.return_value
        mock_resource_maker.assert_called_once_with()
        assert bucket.mapping[hikari.Snowflake(123)] is mock_resource_maker.return_value

    def test_cleanup(self):
        mock_cooldown_1 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_cooldown_2 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_cooldown_3 = mock.Mock(has_expired=mock.Mock(return_value=False))
        bucket = tanjun.dependencies.limiters._FlatResource(tanjun.BucketResource.USER, mock.Mock())
        bucket.mapping = {
            hikari.Snowflake(123312): mock_cooldown_1,
            hikari.Snowflake(4321123): mock.Mock(has_expired=mock.Mock(return_value=True)),
            hikari.Snowflake(54123): mock_cooldown_2,
            hikari.Snowflake(222): mock.Mock(has_expired=mock.Mock(return_value=True)),
            hikari.Snowflake(654124): mock_cooldown_3,
            hikari.Snowflake(123321): mock.Mock(has_expired=mock.Mock(return_value=True)),
        }

        bucket.cleanup()

        assert bucket.mapping == {
            hikari.Snowflake(123312): mock_cooldown_1,
            hikari.Snowflake(54123): mock_cooldown_2,
            hikari.Snowflake(654124): mock_cooldown_3,
        }

    def test_copy(self):
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._FlatResource(tanjun.BucketResource.PARENT_CHANNEL, mock_resource_maker)
        bucket.mapping[hikari.Snowflake(321123)] = mock.Mock()

        new_bucket = bucket.copy()

        assert new_bucket is not bucket
        assert new_bucket.resource is bucket.resource
        assert new_bucket.make_resource is mock_resource_maker
        assert new_bucket.mapping == {}


class Test_MemberResource:
    @pytest.mark.asyncio()
    async def test_into_inner(self) -> None:
        mock_resource_maker = mock.Mock()
        mock_resource = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.mapping[hikari.Snowflake(654123)] = {hikari.Snowflake(3333): mock_resource}
        mock_context = mock.Mock(guild_id=hikari.Snowflake(654123))
        mock_context.author.id = hikari.Snowflake(3333)

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource
        mock_resource_maker.assert_not_called()
        assert bucket.mapping[hikari.Snowflake(654123)] == {hikari.Snowflake(3333): mock_resource}

    @pytest.mark.asyncio()
    async def test_into_inner_when_member_not_found(self) -> None:
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.mapping[hikari.Snowflake(65234234)] = {hikari.Snowflake(123321): mock.Mock()}
        mock_context = mock.Mock(guild_id=hikari.Snowflake(65234234))
        mock_context.author.id = hikari.Snowflake(123542)

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource_maker.return_value
        mock_resource_maker.assert_called_once_with()
        assert bucket.mapping[hikari.Snowflake(65234234)][hikari.Snowflake(123542)] is mock_resource_maker.return_value

    @pytest.mark.asyncio()
    async def test_into_inner_when_guild_not_found(self) -> None:
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        mock_context = mock.Mock(guild_id=hikari.Snowflake(123123))
        mock_context.author.id = hikari.Snowflake(32212)

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource_maker.return_value
        mock_resource_maker.assert_called_once_with()
        assert bucket.mapping[hikari.Snowflake(123123)] == {hikari.Snowflake(32212): mock_resource_maker.return_value}

    @pytest.mark.asyncio()
    async def test_into_inner_when_dm_bound(self) -> None:
        mock_resource_maker = mock.Mock()
        mock_resource = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.dm_fallback[hikari.Snowflake(1233214)] = mock_resource
        mock_context = mock.Mock(guild_id=None, channel_id=hikari.Snowflake(1233214))

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource
        mock_resource_maker.assert_not_called()
        assert bucket.dm_fallback[hikari.Snowflake(1233214)] is mock_resource

    @pytest.mark.asyncio()
    async def test_into_inner_when_dm_bound_and_user_not_found(self) -> None:
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        mock_context = mock.Mock(guild_id=None, channel_id=hikari.Snowflake(123312))

        result = await bucket.into_inner(mock_context)

        assert result is mock_resource_maker.return_value
        mock_resource_maker.assert_called_once_with()
        assert bucket.dm_fallback[hikari.Snowflake(123312)] is mock_resource_maker.return_value

    @pytest.mark.asyncio()
    async def test_try_into_inner(self) -> None:
        mock_resource_maker = mock.Mock()
        mock_resource = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.mapping[hikari.Snowflake(222222)] = {hikari.Snowflake(666666): mock_resource}
        mock_context = mock.Mock(guild_id=hikari.Snowflake(222222))
        mock_context.author.id = hikari.Snowflake(666666)

        result = await bucket.try_into_inner(mock_context)

        assert result is mock_resource
        mock_resource_maker.assert_not_called()
        assert bucket.mapping[hikari.Snowflake(222222)] == {hikari.Snowflake(666666): mock_resource}

    @pytest.mark.asyncio()
    async def test_try_into_inner_when_member_not_found(self) -> None:
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.mapping[hikari.Snowflake(652134)] = {hikari.Snowflake(123321): mock.Mock()}
        mock_context = mock.Mock(guild_id=hikari.Snowflake(652134))
        mock_context.author.id = hikari.Snowflake(43213)

        result = await bucket.try_into_inner(mock_context)

        assert result is None
        mock_resource_maker.assert_not_called()
        assert hikari.Snowflake(43213) not in bucket.mapping[hikari.Snowflake(652134)]

    @pytest.mark.asyncio()
    async def test_try_into_inner_when_guild_not_found(self) -> None:
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        mock_context = mock.Mock(guild_id=hikari.Snowflake(4234))
        mock_context.author.id = hikari.Snowflake(123321)

        result = await bucket.try_into_inner(mock_context)

        assert result is None
        mock_resource_maker.assert_not_called()
        assert hikari.Snowflake(4234) not in bucket.mapping

    @pytest.mark.asyncio()
    async def test_try_into_inner_when_dm_bound(self) -> None:
        mock_resource_maker = mock.Mock()
        mock_resource = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.dm_fallback[hikari.Snowflake(76345)] = mock_resource
        mock_context = mock.Mock(guild_id=None, channel_id=hikari.Snowflake(76345))

        result = await bucket.try_into_inner(mock_context)

        assert result is mock_resource
        mock_resource_maker.assert_not_called()
        assert bucket.dm_fallback[hikari.Snowflake(76345)] is mock_resource

    @pytest.mark.asyncio()
    async def test_try_into_inner_when_dm_bound_and_user_not_found(self) -> None:
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        mock_context = mock.Mock(guild_id=None, channel_id=hikari.Snowflake(555555))

        result = await bucket.try_into_inner(mock_context)

        assert result is None
        mock_resource_maker.assert_not_called()
        assert hikari.Snowflake(555555) not in bucket.dm_fallback

    def test_cleanup(self):
        mock_cooldown_1 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_cooldown_2 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_cooldown_3 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_dm_cooldown_1 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_dm_cooldown_2 = mock.Mock(has_expired=mock.Mock(return_value=False))
        mock_dm_cooldown_3 = mock.Mock(has_expired=mock.Mock(return_value=False))
        bucket = tanjun.dependencies.limiters._MemberResource(mock.Mock())
        bucket.mapping = {
            hikari.Snowflake(54123): {
                hikari.Snowflake(123312): mock_cooldown_1,
                hikari.Snowflake(4321123): mock.Mock(has_expired=mock.Mock(return_value=True)),
                hikari.Snowflake(54123): mock_cooldown_2,
            },
            hikari.Snowflake(666): {
                hikari.Snowflake(4321123): mock.Mock(has_expired=mock.Mock(return_value=True)),
            },
            hikari.Snowflake(6512312): {
                hikari.Snowflake(222): mock.Mock(has_expired=mock.Mock(return_value=True)),
                hikari.Snowflake(654124): mock_cooldown_3,
                hikari.Snowflake(123321): mock.Mock(has_expired=mock.Mock(return_value=True)),
            },
        }
        bucket.dm_fallback = {
            hikari.Snowflake(231123): mock.Mock(has_expired=mock.Mock(return_value=True)),
            hikari.Snowflake(696969): mock_dm_cooldown_1,
            hikari.Snowflake(3214312): mock.Mock(has_expired=mock.Mock(return_value=True)),
            hikari.Snowflake(969696): mock_dm_cooldown_2,
            hikari.Snowflake(123321123): mock.Mock(has_expired=mock.Mock(return_value=True)),
            hikari.Snowflake(42069): mock_dm_cooldown_3,
        }

        bucket.cleanup()

        assert bucket.mapping == {
            hikari.Snowflake(54123): {
                hikari.Snowflake(123312): mock_cooldown_1,
                hikari.Snowflake(54123): mock_cooldown_2,
            },
            hikari.Snowflake(6512312): {hikari.Snowflake(654124): mock_cooldown_3},
        }
        assert bucket.dm_fallback == {
            hikari.Snowflake(696969): mock_dm_cooldown_1,
            hikari.Snowflake(969696): mock_dm_cooldown_2,
            hikari.Snowflake(42069): mock_dm_cooldown_3,
        }

    def test_copy(self):
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._MemberResource(mock_resource_maker)
        bucket.mapping[hikari.Snowflake(321123)] = mock.Mock()
        bucket.dm_fallback[hikari.Snowflake(541123)] = mock.Mock()

        new_bucket = bucket.copy()

        assert new_bucket is not bucket
        assert new_bucket.make_resource is mock_resource_maker
        assert new_bucket.mapping == {}
        assert new_bucket.dm_fallback == {}


class Test_GlobalResource:
    @pytest.mark.asyncio()
    async def test_into_inner(self):
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._GlobalResource(mock_resource_maker)

        assert await bucket.into_inner(mock.Mock()) is mock_resource_maker.return_value
        assert await bucket.into_inner(mock.Mock()) is mock_resource_maker.return_value
        mock_resource_maker.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_try_into_inner(self):
        mock_resource_maker = mock.Mock()
        bucket = tanjun.dependencies.limiters._GlobalResource(mock_resource_maker)

        assert await bucket.try_into_inner(mock.Mock()) is mock_resource_maker.return_value
        assert await bucket.try_into_inner(mock.Mock()) is mock_resource_maker.return_value
        mock_resource_maker.assert_called_once_with()

    def test_cleanup(self):
        tanjun.dependencies.limiters._GlobalResource(mock.Mock()).cleanup()

    def test_copy(self):
        mock_resource_1 = mock.Mock()
        mock_resource_2 = mock.Mock()
        mock_resource_maker = mock.Mock(side_effect=[mock_resource_1, mock_resource_2])
        bucket = tanjun.dependencies.limiters._GlobalResource(mock_resource_maker)
        bucket.bucket.increment()

        new_bucket = bucket.copy()

        assert new_bucket is not bucket
        assert new_bucket.make_resource is mock_resource_maker
        assert new_bucket.bucket is mock_resource_2
        assert new_bucket.make_resource is mock_resource_maker
        mock_resource_maker.assert_has_calls([mock.call(), mock.call()])

        assert bucket.bucket is mock_resource_1


class TestInMemoryCooldownManager:
    @pytest.mark.asyncio()
    async def test__gc(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_bucket_1 = mock.Mock()
        mock_bucket_2 = mock.Mock()
        mock_bucket_3 = mock.Mock()
        manager._buckets = {"e": mock_bucket_1, "a": mock_bucket_2, "f": mock_bucket_3}
        mock_error = Exception("test")

        with mock.patch.object(asyncio, "sleep", side_effect=[None, None, mock_error]) as sleep:
            with pytest.raises(Exception) as exc_info:  # noqa: PT011, PT012
                await asyncio.wait_for(manager._gc(), timeout=0.5)

                assert exc_info.value is mock_error
                sleep.assert_has_awaits([mock.call(10), mock.call(10), mock.call(10)])

        mock_bucket_1.cleanup.assert_has_calls([mock.call(), mock.call()])
        mock_bucket_2.cleanup.assert_has_calls([mock.call(), mock.call()])
        mock_bucket_3.cleanup.assert_has_calls([mock.call(), mock.call()])

    def test_add_to_client(self):
        mock_client = mock.Mock(tanjun.Client, is_alive=False)
        mock_open = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryCooldownManager):
            open = mock_open

        manager = StubManager()
        manager.add_to_client(mock_client)

        mock_client.add_client_callback.assert_has_calls(
            [
                mock.call(tanjun.abc.ClientCallbackNames.STARTING, manager.open),
                mock.call(tanjun.abc.ClientCallbackNames.CLOSING, manager.close),
            ]
        )
        mock_open.assert_not_called()

    def test_add_to_client_when_client_is_active(self):
        mock_client = mock.Mock(tanjun.Client, is_alive=True)
        mock_open = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryCooldownManager):
            open = mock_open

        manager = StubManager()
        manager.add_to_client(mock_client)

        mock_client.add_client_callback.assert_has_calls(
            [
                mock.call(tanjun.abc.ClientCallbackNames.STARTING, manager.open),
                mock.call(tanjun.abc.ClientCallbackNames.CLOSING, manager.close),
            ]
        )
        mock_open.assert_called_once_with(_loop=mock_client.loop)

    @pytest.mark.asyncio()
    async def test_check_cooldown(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock(try_into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.try_into_inner.return_value
        manager._buckets["ok"] = mock_bucket

        result = await manager.check_cooldown("ok", mock_context, increment=False)

        assert result is mock_bucket.try_into_inner.return_value.must_wait_for.return_value
        mock_bucket.try_into_inner.assert_awaited_once_with(mock_context)
        mock_inner.must_wait_for.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_check_cooldown_when_try_into_inner_returns_none(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock(try_into_inner=mock.AsyncMock(return_value=None))
        manager._buckets["ok"] = mock_bucket

        result = await manager.check_cooldown("ok", mock_context, increment=False)

        assert result is None
        mock_bucket.try_into_inner.assert_awaited_once_with(mock_context)

    @pytest.mark.asyncio()
    async def test_check_cooldown_when_increment(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.into_inner.return_value
        mock_inner.increment.return_value = mock_bucket.into_inner.return_value
        mock_inner.must_wait_for.return_value = None
        manager._buckets["ok"] = mock_bucket

        result = await manager.check_cooldown("ok", mock_context, increment=True)

        assert result is None
        mock_bucket.into_inner.assert_awaited_once_with(mock_context)
        mock_inner.increment.assert_called_once_with()
        mock_inner.must_wait_for.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_check_cooldown_when_increment_and_must_wait_for_returns_time(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.into_inner.return_value
        mock_inner.increment.return_value = mock_inner
        manager._buckets["ok"] = mock_bucket

        result = await manager.check_cooldown("ok", mock_context, increment=True)

        assert result is mock_inner.must_wait_for.return_value
        mock_bucket.into_inner.assert_awaited_once_with(mock_context)
        mock_inner.increment.assert_not_called()
        mock_inner.must_wait_for.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_check_cooldown_falls_back_to_default_when_increment_creates_new_bucket(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock()
        mock_bucket.copy.return_value = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.copy.return_value.into_inner.return_value
        mock_inner.increment.return_value = mock_inner
        mock_inner.must_wait_for.return_value = None

        manager._default_bucket_template = mock_bucket

        result = await manager.check_cooldown("ok", mock_context, increment=True)

        assert result is None
        assert manager._buckets["ok"] is mock_bucket.copy.return_value
        mock_bucket.copy.assert_called_once_with()
        mock_bucket.copy.return_value.into_inner.assert_awaited_once_with(mock_context)
        mock_inner.increment.assert_called_once_with()
        mock_inner.must_wait_for.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_check_cooldown_falls_back_to_default_when_increment_creates_new_bucket_when_wait_for_returns_time(
        self,
    ):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock()
        mock_bucket.copy.return_value = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.copy.return_value.into_inner.return_value
        mock_inner.increment.return_value = mock_inner

        manager._default_bucket_template = mock_bucket

        result = await manager.check_cooldown("ok", mock_context, increment=True)

        assert result is mock_inner.must_wait_for.return_value
        assert manager._buckets["ok"] is mock_bucket.copy.return_value
        mock_bucket.copy.assert_called_once_with()
        mock_bucket.copy.return_value.into_inner.assert_awaited_once_with(mock_context)
        mock_inner.increment.assert_not_called()
        mock_inner.must_wait_for.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_increment_cooldown(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.into_inner.return_value
        manager._buckets["catgirl neko"] = mock_bucket

        result = await manager.increment_cooldown("catgirl neko", mock_context)

        assert result is None
        mock_bucket.into_inner.assert_awaited_once_with(mock_context)
        mock_inner.increment.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_increment_cooldown_falls_back_to_default(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_context = mock.Mock()
        mock_bucket = mock.Mock()
        mock_bucket.copy.return_value.into_inner = mock.AsyncMock(return_value=mock.Mock())
        mock_inner = mock_bucket.copy.return_value.into_inner.return_value
        manager._default_bucket_template = mock_bucket

        result = await manager.increment_cooldown("69", mock_context)

        assert result is None
        assert manager._buckets["69"] is mock_bucket.copy.return_value
        mock_bucket.copy.assert_called_once_with()
        mock_bucket.copy.return_value.into_inner.assert_awaited_once_with(mock_context)
        mock_inner.increment.assert_called_once_with()

    def test_close(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        mock_gc_task = mock.Mock()
        manager._gc_task = mock_gc_task

        manager.close()

        mock_gc_task.cancel.assert_called_once_with()

    def test_close_when_not_active(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with pytest.raises(RuntimeError, match="Cooldown manager is not active"):
            manager.close()

    def test_open(self):
        mock_gc = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryCooldownManager):
            _gc = mock_gc

        manager = StubManager()

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            manager.open()

            assert manager._gc_task is get_running_loop.return_value.create_task.return_value
            get_running_loop.assert_called_once_with()
            get_running_loop.return_value.create_task.assert_called_once_with(mock_gc.return_value)
            mock_gc.assert_called_once_with()

    def test_open_with_passed_through_loop(self):
        mock_gc = mock.Mock()
        mock_loop = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryCooldownManager):
            _gc = mock_gc

        manager = StubManager()

        manager.open(_loop=mock_loop)

        assert manager._gc_task is mock_loop.create_task.return_value
        mock_loop.create_task.assert_called_once_with(mock_gc.return_value)
        mock_gc.assert_called_once_with()

    def test_open_when_already_active(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        manager._gc_task = mock.Mock()

        with pytest.raises(RuntimeError, match="Cooldown manager is already running"):
            manager.open()

    def test_disable_bucket(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        default_bucket = manager._default_bucket_template

        with mock.patch.object(tanjun.dependencies.limiters, "_GlobalResource") as cooldown_bucket:
            result = manager.disable_bucket("kitten")

            assert result is manager
            assert manager._buckets["kitten"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is default_bucket
            cooldown_bucket.return_value.copy.assert_not_called()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == -1
            assert cooldown.reset_after == -1

    def test_disable_bucket_when_is_default(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with mock.patch.object(tanjun.dependencies.limiters, "_GlobalResource") as cooldown_bucket:
            result = manager.disable_bucket("default")

            assert result is manager
            assert manager._buckets["default"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is cooldown_bucket.return_value.copy.return_value
            cooldown_bucket.return_value.copy.assert_called_once_with()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == -1
            assert cooldown.reset_after == -1

    @pytest.mark.parametrize(
        "resource_type",
        [
            tanjun.BucketResource.USER,
            tanjun.BucketResource.CHANNEL,
            tanjun.BucketResource.PARENT_CHANNEL,
            tanjun.BucketResource.TOP_ROLE,
            tanjun.BucketResource.GUILD,
        ],
    )
    def test_set_bucket(self, resource_type: tanjun.BucketResource):
        manager = tanjun.dependencies.InMemoryCooldownManager()
        default_bucket = manager._default_bucket_template

        with mock.patch.object(tanjun.dependencies.limiters, "_FlatResource") as cooldown_bucket:
            result = manager.set_bucket("gay catgirl", resource_type, 123, 43.123)

            assert result is manager
            assert manager._buckets["gay catgirl"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is default_bucket
            cooldown_bucket.return_value.copy.assert_not_called()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 2
            assert len(cooldown_bucket.call_args.kwargs) == 0
            assert cooldown_bucket.call_args.args[0] is resource_type

            cooldown_maker = cooldown_bucket.call_args.args[1]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == 123
            assert cooldown.reset_after == 43.123

    @pytest.mark.parametrize("reset_after", [datetime.timedelta(seconds=69), 69, 69.0])
    def test_set_bucket_handles_different_reset_after_types(
        self, reset_after: typing.Union[datetime.timedelta, int, float]
    ):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with mock.patch.object(tanjun.dependencies.limiters, "_FlatResource") as cooldown_bucket:
            result = manager.set_bucket("gay catgirl", tanjun.BucketResource.USER, 444, reset_after)

            assert result is manager
            assert manager._buckets["gay catgirl"] is cooldown_bucket.return_value
            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 2
            assert len(cooldown_bucket.call_args.kwargs) == 0
            assert cooldown_bucket.call_args.args[0] is tanjun.BucketResource.USER

            cooldown_maker = cooldown_bucket.call_args.args[1]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == 444
            assert cooldown.reset_after == 69

    def test_set_bucket_for_member_resource(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with mock.patch.object(tanjun.dependencies.limiters, "_MemberResource") as cooldown_bucket:
            result = manager.set_bucket("meowth", tanjun.BucketResource.MEMBER, 64, 42.0)

            assert result is manager
            assert manager._buckets["meowth"] is cooldown_bucket.return_value
            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == 64
            assert cooldown.reset_after == 42.0

    def test_set_bucket_for_global_resource(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with mock.patch.object(tanjun.dependencies.limiters, "_GlobalResource") as cooldown_bucket:
            result = manager.set_bucket("meow", tanjun.BucketResource.GLOBAL, 420, 69.420)

            assert result is manager
            assert manager._buckets["meow"] is cooldown_bucket.return_value
            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == 420
            assert cooldown.reset_after == 69.420

    def test_set_bucket_when_is_default(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with mock.patch.object(tanjun.dependencies.limiters, "_FlatResource") as cooldown_bucket:
            result = manager.set_bucket("default", tanjun.BucketResource.USER, 777, 666.0)

            assert result is manager
            assert manager._buckets["default"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is cooldown_bucket.return_value.copy.return_value
            cooldown_bucket.return_value.copy.assert_called_once_with()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 2
            assert len(cooldown_bucket.call_args.kwargs) == 0
            assert cooldown_bucket.call_args.args[0] is tanjun.BucketResource.USER

            cooldown_maker = cooldown_bucket.call_args.args[1]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._Cooldown)
            assert cooldown.limit == 777
            assert cooldown.reset_after == 666.0

    @pytest.mark.parametrize("reset_after", [datetime.timedelta(seconds=-42), -431, -0.123])
    def test_set_bucket_when_reset_after_is_negative(self, reset_after: typing.Union[datetime.timedelta, float, int]):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with pytest.raises(ValueError, match="reset_after must be greater than 0 seconds"):
            manager.set_bucket("gay catgirl", tanjun.BucketResource.USER, 123, reset_after)

    def test_set_bucket_when_limit_is_negative(self):
        manager = tanjun.dependencies.InMemoryCooldownManager()

        with pytest.raises(ValueError, match="limit must be greater than 0"):
            manager.set_bucket("gay catgirl", tanjun.BucketResource.USER, -123, 43.123)


class TestCooldownPreExecution:
    @pytest.mark.asyncio()
    async def test_call(self):
        pre_execution = tanjun.dependencies.CooldownPreExecution("yuri catgirls", owners_exempt=False)
        mock_context = mock.Mock()
        mock_cooldown_manager = mock.AsyncMock()
        mock_cooldown_manager.check_cooldown.return_value = None
        mock_owner_check = mock.AsyncMock()

        await pre_execution(mock_context, cooldowns=mock_cooldown_manager, owner_check=mock_owner_check)

        mock_cooldown_manager.check_cooldown.assert_awaited_once_with("yuri catgirls", mock_context, increment=True)
        mock_owner_check.check_ownership.assert_not_called()

    @pytest.mark.asyncio()
    async def test_call_when_owners_exempt(self):
        pre_execution = tanjun.dependencies.CooldownPreExecution("yuri catgirls", owners_exempt=True)
        mock_context = mock.Mock()
        mock_cooldown_manager = mock.AsyncMock()
        mock_owner_check = mock.AsyncMock()
        mock_owner_check.check_ownership.return_value = True

        await pre_execution(mock_context, cooldowns=mock_cooldown_manager, owner_check=mock_owner_check)

        mock_cooldown_manager.check_cooldown.assert_not_called()
        mock_owner_check.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)

    @pytest.mark.asyncio()
    async def test_call_when_owners_exempt_but_owner_check_is_none(self):
        pre_execution = tanjun.dependencies.CooldownPreExecution("yuri catgirls", owners_exempt=True)
        mock_context = mock.Mock()
        mock_cooldown_manager = mock.AsyncMock()
        mock_cooldown_manager.check_cooldown.return_value = None

        await pre_execution(mock_context, cooldowns=mock_cooldown_manager, owner_check=None)

        mock_cooldown_manager.check_cooldown.assert_called_once_with("yuri catgirls", mock_context, increment=True)
        assert pre_execution._owners_exempt is False

    @pytest.mark.asyncio()
    async def test_call_when_owners_exempt_and_not_owner(self):
        pre_execution = tanjun.dependencies.CooldownPreExecution("catgirls", owners_exempt=True)
        mock_context = mock.Mock()
        mock_cooldown_manager = mock.AsyncMock()
        mock_cooldown_manager.check_cooldown.return_value = None
        mock_owner_check = mock.AsyncMock()
        mock_owner_check.check_ownership.return_value = False

        await pre_execution(mock_context, cooldowns=mock_cooldown_manager, owner_check=mock_owner_check)

        mock_cooldown_manager.check_cooldown.assert_awaited_once_with("catgirls", mock_context, increment=True)
        mock_owner_check.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)

    @pytest.mark.asyncio()
    async def test_call_when_owners_exempt_still_leads_to_wait_for(self):
        pre_execution = tanjun.dependencies.CooldownPreExecution("yuri", owners_exempt=True)
        mock_context = mock.Mock()
        mock_cooldown_manager = mock.AsyncMock()
        mock_cooldown_manager.check_cooldown.return_value = 69.420
        mock_owner_check = mock.AsyncMock()
        mock_owner_check.check_ownership.return_value = False

        with pytest.raises(tanjun.CommandError, match="Please wait 69.42 seconds before using this command again."):
            await pre_execution(mock_context, cooldowns=mock_cooldown_manager, owner_check=mock_owner_check)

        mock_cooldown_manager.check_cooldown.assert_awaited_once_with("yuri", mock_context, increment=True)
        mock_owner_check.check_ownership.assert_awaited_once_with(mock_context.client, mock_context.author)

    @pytest.mark.asyncio()
    async def test_call_when_wait_for(self):
        pre_execution = tanjun.dependencies.CooldownPreExecution("catgirls yuri", owners_exempt=False)
        mock_context = mock.Mock()
        mock_cooldown_manager = mock.AsyncMock()
        mock_cooldown_manager.check_cooldown.return_value = 420.69420
        mock_owner_check = mock.AsyncMock()

        with pytest.raises(tanjun.CommandError, match="Please wait 420.69 seconds before using this command again."):
            await pre_execution(mock_context, cooldowns=mock_cooldown_manager, owner_check=mock_owner_check)

        mock_cooldown_manager.check_cooldown.assert_awaited_once_with("catgirls yuri", mock_context, increment=True)
        mock_owner_check.check_ownership.assert_not_called()


def test_with_cooldown():
    mock_command = mock.Mock()

    with mock.patch.object(tanjun.dependencies.limiters, "CooldownPreExecution") as mock_pre_execution:
        tanjun.with_cooldown("catgirl x catgirl", error_message="pussy cat pussy cat", owners_exempt=False)(
            mock_command
        )

        mock_pre_execution.assert_called_once_with(
            "catgirl x catgirl", error_message="pussy cat pussy cat", owners_exempt=False
        )
        mock_command.hooks.add_pre_execution.assert_called_once_with(mock_pre_execution.return_value)


def test_with_cooldown_when_no_set_hooks():
    mock_command = mock.Mock(hooks=None)

    stack = contextlib.ExitStack()
    mock_pre_execution = stack.enter_context(mock.patch.object(tanjun.dependencies.limiters, "CooldownPreExecution"))
    any_hooks = stack.enter_context(mock.patch.object(tanjun.hooks, "AnyHooks"))

    with stack:
        tanjun.with_cooldown("catgirl x catgirl", error_message="pussy cat pussy cat", owners_exempt=False)(
            mock_command
        )

        mock_pre_execution.assert_called_once_with(
            "catgirl x catgirl", error_message="pussy cat pussy cat", owners_exempt=False
        )
        mock_command.set_hooks.assert_called_once_with(any_hooks.return_value)
        any_hooks.return_value.add_pre_execution.assert_called_once_with(mock_pre_execution.return_value)
        any_hooks.assert_called_once_with()


class Test_ConcurrencyLimit:
    def test_acquire(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(2)

        result = limit.acquire()

        assert result is True
        assert limit.counter == 1

    def test_acquire_when_couldnt_acquire(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(2)
        limit.counter = 2

        result = limit.acquire()

        assert result is False
        assert limit.counter == 2

    def test_acquire_when_limit_is_negative_1(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(-1)
        limit.counter = 2

        result = limit.acquire()

        assert result is True
        assert limit.counter == 2

    def test_release(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(2)
        limit.counter = 2

        limit.release()

        assert limit.counter == 1

    def test_release_when_not_acquired(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(2)

        with pytest.raises(RuntimeError, match="Cannot release a limit that has not been acquired"):
            limit.release()

    def test_release_when_limit_is_negative_1(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(-1)
        limit.counter = 0

        limit.release()

        assert limit.counter == 0

    def test_has_expired(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(2)
        limit.counter = 1

        assert limit.has_expired() is False

    def test_has_expired_when_has_expired(self):
        limit = tanjun.dependencies.limiters._ConcurrencyLimit(2)
        limit.counter = 0

        assert limit.has_expired() is True


class TestInMemoryConcurrencyLimiter:
    @pytest.mark.asyncio()
    async def test__gc(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        mock_bucket_1 = mock.Mock()
        mock_bucket_2 = mock.Mock()
        mock_bucket_3 = mock.Mock()
        manager._buckets = {"e": mock_bucket_1, "a": mock_bucket_2, "f": mock_bucket_3}
        mock_error = Exception("test")

        with mock.patch.object(asyncio, "sleep", side_effect=[None, None, mock_error]) as sleep:
            with pytest.raises(Exception) as exc_info:  # noqa: PT011, PT012
                await asyncio.wait_for(manager._gc(), timeout=0.5)

                assert exc_info.value is mock_error
                sleep.assert_has_awaits([mock.call(10), mock.call(10), mock.call(10)])

        mock_bucket_1.cleanup.assert_has_calls([mock.call(), mock.call()])
        mock_bucket_2.cleanup.assert_has_calls([mock.call(), mock.call()])
        mock_bucket_3.cleanup.assert_has_calls([mock.call(), mock.call()])

    def test_add_to_client(self):
        mock_client = mock.Mock(tanjun.Client, is_alive=False)
        mock_open = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryConcurrencyLimiter):
            open = mock_open

        manager = StubManager()
        manager.add_to_client(mock_client)

        mock_client.add_client_callback.assert_has_calls(
            [
                mock.call(tanjun.abc.ClientCallbackNames.STARTING, manager.open),
                mock.call(tanjun.abc.ClientCallbackNames.CLOSING, manager.close),
            ]
        )
        mock_open.assert_not_called()

    def test_add_to_client_when_client_is_active(self):
        mock_client = mock.Mock(tanjun.Client, is_alive=True)
        mock_open = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryConcurrencyLimiter):
            open = mock_open

        manager = StubManager()
        manager.add_to_client(mock_client)

        mock_client.add_client_callback.assert_has_calls(
            [
                mock.call(tanjun.abc.ClientCallbackNames.STARTING, manager.open),
                mock.call(tanjun.abc.ClientCallbackNames.CLOSING, manager.close),
            ]
        )
        mock_open.assert_called_once_with(_loop=mock_client.loop)

    def test_close(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        mock_gc_task = mock.Mock()
        manager._gc_task = mock_gc_task

        manager.close()

        mock_gc_task.cancel.assert_called_once_with()

    def test_close_when_not_active(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()

        with pytest.raises(RuntimeError, match="Concurrency manager is not active"):
            manager.close()

    def test_open(self):
        mock_gc = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryConcurrencyLimiter):
            _gc = mock_gc

        manager = StubManager()

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            manager.open()

            assert manager._gc_task is get_running_loop.return_value.create_task.return_value
            get_running_loop.assert_called_once_with()
            get_running_loop.return_value.create_task.assert_called_once_with(mock_gc.return_value)
            mock_gc.assert_called_once_with()

    def test_open_with_passed_through_loop(self):
        mock_gc = mock.Mock()
        mock_loop = mock.Mock()

        class StubManager(tanjun.dependencies.InMemoryConcurrencyLimiter):
            _gc = mock_gc

        manager = StubManager()

        manager.open(_loop=mock_loop)

        assert manager._gc_task is mock_loop.create_task.return_value
        mock_loop.create_task.assert_called_once_with(mock_gc.return_value)
        mock_gc.assert_called_once_with()

    def test_open_when_already_active(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        manager._gc_task = mock.Mock()

        with pytest.raises(RuntimeError, match="Concurrency manager is already running"):
            manager.open()

    @pytest.mark.asyncio()
    async def test_try_acquire(self):
        mock_bucket = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.into_inner.return_value
        mock_inner.acquire.return_value = True
        mock_context = mock.Mock()
        manager = tanjun.InMemoryConcurrencyLimiter()
        manager._buckets["aye"] = mock_bucket

        result = await manager.try_acquire("aye", mock_context)

        assert result is True
        mock_bucket.into_inner.assert_called_once_with(mock_context)
        mock_inner.acquire.assert_called_once_with()
        assert manager._acquiring_ctxs[("aye", mock_context)] is mock_inner

    @pytest.mark.asyncio()
    async def test_try_acquire_when_failed_to_acquire(self):
        mock_bucket = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.into_inner.return_value
        mock_inner.acquire.return_value = False
        mock_context = mock.Mock()
        manager = tanjun.InMemoryConcurrencyLimiter()
        manager._buckets["nya"] = mock_bucket

        result = await manager.try_acquire("nya", mock_context)

        assert result is False
        mock_bucket.into_inner.assert_called_once_with(mock_context)
        mock_inner.acquire.assert_called_once_with()
        assert ("nya", mock_context) not in manager._acquiring_ctxs

    @pytest.mark.asyncio()
    async def test_try_acquire_for_already_acquired_context(self):
        mock_bucket = mock.Mock()
        mock_context = mock.Mock()
        mock_limiter = mock.Mock()
        manager = tanjun.InMemoryConcurrencyLimiter()
        manager._buckets["ayee"] = mock_bucket
        manager._acquiring_ctxs[("ayee", mock_context)] = mock_limiter

        result = await manager.try_acquire("ayee", mock_context)

        assert result is True
        mock_bucket.into_inner.assert_not_called()
        mock_limiter.acquire.assert_not_called()
        assert manager._acquiring_ctxs[("ayee", mock_context)] is mock_limiter

    @pytest.mark.asyncio()
    async def test_try_acquire_falls_back_to_default_bucket(self):
        mock_bucket = mock.Mock(into_inner=mock.AsyncMock(return_value=mock.Mock()))
        mock_inner: typing.Any = mock_bucket.into_inner.return_value
        mock_inner.acquire.return_value = True
        mock_bucket_template = mock.Mock()
        mock_bucket_template.copy.return_value = mock_bucket
        mock_context = mock.Mock()
        manager = tanjun.InMemoryConcurrencyLimiter()
        manager._default_bucket_template = mock_bucket_template

        result = await manager.try_acquire("yeet", mock_context)

        assert result is True
        mock_bucket_template.copy.assert_called_once_with()
        mock_bucket.into_inner.assert_called_once_with(mock_context)
        mock_inner.acquire.assert_called_once_with()
        assert manager._acquiring_ctxs[("yeet", mock_context)] is mock_inner
        assert manager._buckets["yeet"] is mock_bucket

    @pytest.mark.asyncio()
    async def test_release(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        mock_context = mock.Mock()
        mock_limiter = mock.Mock()
        manager._acquiring_ctxs[("nya", mock_context)] = mock_limiter

        await manager.release("nya", mock_context)

        assert ("nya", mock_context) not in manager._acquiring_ctxs
        mock_limiter.release.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_release_for_unknown_context(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        mock_context = mock.Mock()

        await manager.release("meow", mock_context)

    def test_disable_bucket(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        default_bucket = manager._default_bucket_template

        with mock.patch.object(tanjun.dependencies.limiters, "_GlobalResource") as cooldown_bucket:
            result = manager.disable_bucket("kitty cat")

            assert result is manager
            assert manager._buckets["kitty cat"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is default_bucket
            cooldown_bucket.return_value.copy.assert_not_called()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._ConcurrencyLimit)
            assert cooldown.limit == -1

    def test_disable_bucket_when_is_default(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()

        with mock.patch.object(tanjun.dependencies.limiters, "_GlobalResource") as cooldown_bucket:
            result = manager.disable_bucket("default")

            assert result is manager
            assert manager._buckets["default"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is cooldown_bucket.return_value.copy.return_value
            cooldown_bucket.return_value.copy.assert_called_once_with()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._ConcurrencyLimit)
            assert cooldown.limit == -1

    @pytest.mark.parametrize(
        "resource_type",
        [
            tanjun.BucketResource.USER,
            tanjun.BucketResource.CHANNEL,
            tanjun.BucketResource.PARENT_CHANNEL,
            tanjun.BucketResource.TOP_ROLE,
            tanjun.BucketResource.GUILD,
        ],
    )
    def test_set_bucket(self, resource_type: tanjun.BucketResource):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()
        default_bucket = manager._default_bucket_template

        with mock.patch.object(tanjun.dependencies.limiters, "_FlatResource") as cooldown_bucket:
            result = manager.set_bucket("gay catgirl", resource_type, 321)

            assert result is manager
            assert manager._buckets["gay catgirl"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is default_bucket
            cooldown_bucket.return_value.copy.assert_not_called()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 2
            assert len(cooldown_bucket.call_args.kwargs) == 0
            assert cooldown_bucket.call_args.args[0] is resource_type

            cooldown_maker = cooldown_bucket.call_args.args[1]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._ConcurrencyLimit)
            assert cooldown.limit == 321

    def test_set_bucket_for_member_resource(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()

        with mock.patch.object(tanjun.dependencies.limiters, "_MemberResource") as cooldown_bucket:
            result = manager.set_bucket("meowth", tanjun.BucketResource.MEMBER, 69)

            assert result is manager
            assert manager._buckets["meowth"] is cooldown_bucket.return_value
            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._ConcurrencyLimit)
            assert cooldown.limit == 69

    def test_set_bucket_for_global_resource(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()

        with mock.patch.object(tanjun.dependencies.limiters, "_GlobalResource") as cooldown_bucket:
            result = manager.set_bucket("meow", tanjun.BucketResource.GLOBAL, 42069)

            assert result is manager
            assert manager._buckets["meow"] is cooldown_bucket.return_value
            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 1
            assert len(cooldown_bucket.call_args.kwargs) == 0

            cooldown_maker = cooldown_bucket.call_args.args[0]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._ConcurrencyLimit)
            assert cooldown.limit == 42069

    def test_set_bucket_when_is_default(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()

        with mock.patch.object(tanjun.dependencies.limiters, "_FlatResource") as cooldown_bucket:
            result = manager.set_bucket("default", tanjun.BucketResource.USER, 697)

            assert result is manager
            assert manager._buckets["default"] is cooldown_bucket.return_value
            assert manager._default_bucket_template is cooldown_bucket.return_value.copy.return_value
            cooldown_bucket.return_value.copy.assert_called_once_with()

            assert cooldown_bucket.call_count == 1
            assert len(cooldown_bucket.call_args.args) == 2
            assert len(cooldown_bucket.call_args.kwargs) == 0
            assert cooldown_bucket.call_args.args[0] is tanjun.BucketResource.USER

            cooldown_maker = cooldown_bucket.call_args.args[1]
            cooldown = cooldown_maker()
            assert isinstance(cooldown, tanjun.dependencies.limiters._ConcurrencyLimit)
            assert cooldown.limit == 697

    def test_set_bucket_when_limit_is_negative(self):
        manager = tanjun.dependencies.InMemoryConcurrencyLimiter()

        with pytest.raises(ValueError, match="limit must be greater than 0"):
            manager.set_bucket("gay catgirl", tanjun.BucketResource.USER, -1)


class TestConcurrencyPreExecution:
    @pytest.mark.asyncio()
    async def test_call(self):
        mock_context = mock.Mock()
        mock_limiter = mock.AsyncMock()
        mock_limiter.try_acquire.return_value = True
        hook = tanjun.dependencies.ConcurrencyPreExecution("bucket boobs")

        await hook(mock_context, mock_limiter)

        mock_limiter.try_acquire.assert_awaited_once_with("bucket boobs", mock_context)

    @pytest.mark.asyncio()
    async def test_call_when_acquire_fails(self):
        mock_context = mock.Mock()
        mock_limiter = mock.AsyncMock()
        mock_limiter.try_acquire.return_value = False
        hook = tanjun.dependencies.ConcurrencyPreExecution("bucket catgirls", error_message="an error message")

        with pytest.raises(tanjun.CommandError, match="an error message"):
            await hook(mock_context, mock_limiter)

        mock_limiter.try_acquire.assert_awaited_once_with("bucket catgirls", mock_context)


class TestConcurrencyPostExecution:
    @pytest.mark.asyncio()
    async def test_call(self):
        mock_context = mock.Mock()
        mock_limiter = mock.AsyncMock()
        hook = tanjun.dependencies.ConcurrencyPostExecution("aye bucket")

        await hook(mock_context, mock_limiter)

        mock_limiter.release.assert_awaited_once_with("aye bucket", mock_context)


def test_with_concurrency_limit():
    mock_command = mock.Mock()
    mock_command.hooks.add_pre_execution.return_value = mock_command.hooks
    mock_command.hooks.add_post_execution.return_value = mock_command.hooks
    stack = contextlib.ExitStack()
    pre_execution = stack.enter_context(mock.patch.object(tanjun.dependencies.limiters, "ConcurrencyPreExecution"))
    post_execution = stack.enter_context(mock.patch.object(tanjun.dependencies.limiters, "ConcurrencyPostExecution"))

    result = tanjun.with_concurrency_limit("bucket me", error_message="aye message")(mock_command)

    assert result is mock_command
    mock_command.hooks.add_pre_execution.assert_called_once_with(pre_execution.return_value)
    mock_command.hooks.add_post_execution.assert_called_once_with(post_execution.return_value)
    pre_execution.assert_called_once_with("bucket me", error_message="aye message")
    post_execution.assert_called_once_with("bucket me")


def test_with_concurrency_limit_makes_new_hooks():
    mock_command = mock.Mock(hooks=None)
    stack = contextlib.ExitStack()
    any_hooks = stack.enter_context(mock.patch.object(tanjun.hooks, "AnyHooks"))
    pre_execution = stack.enter_context(mock.patch.object(tanjun.dependencies.limiters, "ConcurrencyPreExecution"))
    post_execution = stack.enter_context(mock.patch.object(tanjun.dependencies.limiters, "ConcurrencyPostExecution"))
    any_hooks.return_value.add_pre_execution.return_value = any_hooks.return_value
    any_hooks.return_value.add_post_execution.return_value = any_hooks.return_value

    result = tanjun.with_concurrency_limit("bucket me", error_message="aye message")(mock_command)

    assert result is mock_command
    any_hooks.assert_called_once_with()
    mock_command.set_hooks.assert_called_once_with(any_hooks.return_value)
    any_hooks.return_value.add_pre_execution.assert_called_once_with(pre_execution.return_value)
    any_hooks.return_value.add_post_execution.assert_called_once_with(post_execution.return_value)
    pre_execution.assert_called_once_with("bucket me", error_message="aye message")
    post_execution.assert_called_once_with("bucket me")
